import os
import plistlib
from typing import IO, Iterable, Optional

from .models import (
    WebBookmarkTypeList,
    WebBookmarkTypeLeaf,
    WebBookmarkType,
    WebBookmarkTypeProxy,
)


class SafariBookmarkItem:
    __slots__ = ("_node", "_parent")

    def __init__(
        self, node: WebBookmarkType, parent: Optional["SafariBookmarkItem"] = None
    ) -> None:
        if not getattr(parent, "is_folder", True):
            raise ValueError("Parent must be a folder")
        self._node = node
        self._parent = parent

    def __str__(self) -> str:
        return self.title

    def __repr__(self) -> str:
        return repr(self._node)

    def __len__(self) -> int:
        return len(self.children)

    def __hash__(self) -> int:
        return hash(self._node)

    def __iter__(self) -> Iterable["SafariBookmarkItem"]:
        for child in getattr(self._node, "children", []):
            yield SafariBookmarkItem(
                node=child,
                parent=self,
            )

    def __contains__(self, other: object) -> bool:
        for child in iter(self):
            if child == other:
                return True
        return False

    def __eq__(self, value: object) -> bool:
        return type(self) is type(value) and self._node == getattr(value, "_node", None)

    def __getitem__(self, key) -> "SafariBookmarkItem":
        if isinstance(key, tuple):
            item = self
            for x in key:
                item = item[x]
            return item
        elif not isinstance(key, str):
            raise TypeError(type(key).__name__)
        for child in iter(self):
            if child.id == key or child.title == key:
                return child
        raise KeyError(key)

    @property
    def movable(self) -> bool:
        return self.is_bookmark or self.is_folder

    @property
    def is_bookmark(self) -> bool:
        return isinstance(self._node, WebBookmarkTypeLeaf)

    @property
    def is_folder(self) -> bool:
        return isinstance(self._node, WebBookmarkTypeList)

    @property
    def type(self) -> str:
        if isinstance(self._node, WebBookmarkTypeProxy):
            return "proxy"
        elif self.is_bookmark:
            return "bookmark"
        elif self.is_folder:
            return "folder"
        else:
            return ""

    @property
    def id(self) -> str:
        return self._node.web_bookmark_uuid

    @property
    def title(self) -> str:
        return getattr(self._node, "title", "")

    @title.setter
    def title(self, title: str) -> None:
        if hasattr(self._node, "title"):
            setattr(self._node, "title", title)

    @property
    def url(self) -> str:
        return getattr(self._node, "url_string", "")

    @url.setter
    def url(self, url: str) -> None:
        if hasattr(self._node, "url_string"):
            setattr(self._node, "url_string", url)

    @property
    def children(self) -> list["SafariBookmarkItem"]:
        return list(iter(self))

    @property
    def parent(self) -> Optional["SafariBookmarkItem"]:
        return self._parent

    def get(self, id: str) -> Optional["SafariBookmarkItem"]:
        if self.id.lower() == id.lower():
            return self
        for child in iter(self):
            result = child.get(id)
            if result is not None:
                return result
        return None

    def walk(self, *titles) -> Optional["SafariBookmarkItem"]:
        if len(titles) == 0:
            return self
        title = titles[0]
        for child in iter(self):
            if child.title == title:
                return child.walk(*titles[1:])
        return None

    def remove(self, item: "SafariBookmarkItem") -> None:
        if item not in self:
            raise RuntimeError("Not a child")
        if not item.movable:
            raise RuntimeError("Invalid item")
        getattr(self._node, "children", []).remove(item._node)
        item._parent = None

    def empty(self) -> None:
        getattr(self._node, "children", []).clear()

    def append(self, item: "SafariBookmarkItem") -> None:
        if not self.is_folder:
            raise RuntimeError("Not a folder")
        if not item.movable:
            raise RuntimeError("Invalid item")
        if item in self:
            return
        if parent := item.parent:
            parent.remove(item)
        getattr(self._node, "children", []).append(item._node)
        item._parent = self

    def add_bookmark(
        self, url: str, title: Optional[str] = None, id: Optional[str] = None
    ) -> "SafariBookmarkItem":
        kwargs = {"URLString": url, "URIDictionary": {}}
        if id is not None:
            kwargs["WebBookmarkUUID"] = id
        if title is not None:
            kwargs["URIDictionary"]["title"] = title
        item = SafariBookmarkItem(node=WebBookmarkTypeLeaf(**kwargs))
        self.append(item)
        return item

    def add_folder(self, title: str, id: Optional[str] = None) -> "SafariBookmarkItem":
        kwargs = {"Title": title, "Children": []}
        if id is not None:
            kwargs["WebBookmarkUUID"] = id
        item = SafariBookmarkItem(node=WebBookmarkTypeList(**kwargs))
        self.append(item)
        return item

    def json(self) -> str:
        return self._node.model_dump_json(by_alias=True)


class SafariBookmarks(SafariBookmarkItem):
    __slots__ = ("_node", "_parent", "_path", "_binary")

    def __init__(self, root: WebBookmarkTypeList) -> None:
        super().__init__(node=root)
        self._path = None
        self._binary = None

    def __enter__(self) -> "SafariBookmarks":
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if exc_type is None and self.path is not None:
            self.save()

    @property
    def movable(self) -> bool:
        return False

    @property
    def path(self) -> Optional[os.PathLike]:
        return self._path

    @property
    def binary(self) -> Optional[bool]:
        return self._binary

    def dump(self, fp: IO, *, binary=True) -> None:
        if hasattr(fp, "mode") and "b" not in fp.mode:
            raise IOError("Must be in binary mode")
        fmt = plistlib.FMT_BINARY if binary else plistlib.FMT_XML
        data = self._node.model_dump(by_alias=True)
        plistlib.dump(data, fp, fmt=fmt, sort_keys=True, skipkeys=False)

    def save(
        self, path: Optional[os.PathLike] = None, binary: Optional[bool] = None
    ) -> None:
        if path is None:
            path = self.path
        if path is None:
            raise RuntimeError("Not opened")
        if binary is None:
            binary = self.binary or True
        with open(path, "wb") as file:
            self.dump(fp=file, binary=binary)

    @classmethod
    def load(cls, fp: IO, *, binary=True) -> "SafariBookmarks":
        if hasattr(fp, "mode") and "b" not in fp.mode:
            raise IOError("Must be in binary mode")
        fmt = plistlib.FMT_BINARY if binary else plistlib.FMT_XML
        data = plistlib.load(fp, fmt=fmt)
        obj = cls(WebBookmarkTypeList.model_validate(data))
        obj._binary = binary
        return obj

    @classmethod
    def open(cls, path: os.PathLike, binary=True) -> "SafariBookmarks":
        with open(path, "rb") as file:
            obj = cls.load(fp=file, binary=binary)
            obj._path = path
            return obj



from pydantic import BaseModel, Field
import uuid
from typing import Annotated, Literal, Union

ChildrenType = Annotated[
    Union["WebBookmarkTypeProxy", "WebBookmarkTypeLeaf", "WebBookmarkTypeList"],
    Field(discriminator="web_bookmark_type"),
]


class CloudKitSync(BaseModel, populate_by_name=True):
    cloud_kit_device_identifier: str = Field(alias="CloudKitDeviceIdentifier")
    cloud_kit_migration_state: int = Field(alias="CloudKitMigrationState")


class WebBookmarkType(BaseModel, extra="allow"):
    web_bookmark_uuid: str = Field(
        alias="WebBookmarkUUID",
        default_factory=lambda: str(uuid.uuid4()),
    )

    def __hash__(self) -> int:
        return hash(self.web_bookmark_uuid)


class WebBookmarkTypeProxy(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeProxy"] = Field(
        alias="WebBookmarkType", default="WebBookmarkTypeProxy"
    )
    title: str = Field(alias="Title")


class WebBookmarkTypeLeaf(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeLeaf"] = Field(
        alias="WebBookmarkType", default="WebBookmarkTypeLeaf"
    )
    url_string: str = Field(alias="URLString")
    uri_dictionary: dict[str, str] = Field(alias="URIDictionary", default_factory=dict)

    @property
    def title(self) -> str:
        return self.uri_dictionary.get("title", "")

    @title.setter
    def title(self, value) -> None:
        self.uri_dictionary["title"] = value


class WebBookmarkTypeList(WebBookmarkType):
    web_bookmark_type: Literal["WebBookmarkTypeList"] = Field(
        alias="WebBookmarkType", default="WebBookmarkTypeList"
    )
    title: str = Field(alias="Title")
    children: list[ChildrenType] = Field(alias="Children", default_factory=list)

    def append(self, item: ChildrenType) -> None:
        self.children.append(item)

    def insert(self, index: int, item: ChildrenType) -> None:
        self.children.insert(index, item)

    def remove(self, item: ChildrenType) -> None:
        self.children.remove(item)

    def empty(self) -> None:
        self.children.clear()



from io import UnsupportedOperation
import os
from typing import IO, List, Optional
from .safaribookmarks import SafariBookmarks, SafariBookmarkItem

DEFAULT_LIST_FORMAT = (
    "{grey}{icon}{reset} {title: <50} {dark_grey}{id: <38}{cyan}{url}{reset}"
)
SIMPLE_FORMAT = "{grey}{icon}{reset} {title: <50} {cyan}{url}{reset}"
ICON_FIRST_LEAF = "┌"
ICON_MIDDLE_LEAF = "├"
ICON_LAST_LEAF = "└"
ICON_SINGLE_LEAF = "─"
ICON_LIST_CONTAINER = "│"

# Source: https://github.com/termcolor/termcolor/blob/main/src/termcolor/termcolor.py
COLORS: dict[str, int] = {
    "reset": 0,
    "black": 30,
    "grey": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "light_grey": 37,
    "dark_grey": 90,
    "light_red": 91,
    "light_green": 92,
    "light_yellow": 93,
    "light_blue": 94,
    "light_magenta": 95,
    "light_cyan": 96,
    "white": 97,
}


class CLI:
    def __init__(self, path: str, out: IO) -> None:
        self.bookmarks = SafariBookmarks.open(path)
        self.output = out
        self.colors = generate_colors(out)

    @property
    def path(self) -> str:
        return str(self.bookmarks.path)

    def run(self, command: str, **kwargs) -> None:
        if command is None:
            raise ValueError("No command specified")
        func = getattr(self, command, None)
        if command.startswith("_") or not callable(func):
            raise ValueError(f"Invalid command: {command}")
        func(**kwargs)

    def _get_or_walk(self, path: List[str]) -> Optional[SafariBookmarkItem]:
        if len(path) == 1:
            result = self.bookmarks.get(path[0])
            if result is not None:
                return result
        return self.bookmarks.walk(*path)

    def _save(self) -> None:
        self.bookmarks.save()

    def _render_item(
        self, item: SafariBookmarkItem, format: str, depth: int = 0, icon: str = ""
    ):
        self.output.write(
            f"{format}\n".format(
                **self.colors,
                icon=icon,
                depth=depth,
                title=item.title.replace("\n", ""),
                type=item.type,
                url=item.url,
                id=item.id,
            )
        )
        if item.is_folder:
            self._render_children(item, format=format, depth=depth + 1)

    def _render_children(self, item: SafariBookmarkItem, format: str, depth: int = 0):
        last_index = len(item) - 1
        for index, child in enumerate(iter(item)):
            icon = ICON_LAST_LEAF if index == last_index else ICON_MIDDLE_LEAF
            if depth == 0 and index == 0:
                icon = ICON_FIRST_LEAF
            if depth == 0 and last_index == 0:
                icon = ICON_SINGLE_LEAF
            if depth > 0:
                icon = f"{ICON_LIST_CONTAINER * depth} {icon}"
            self._render_item(child, format, depth=depth, icon=icon)

    def _render(
        self,
        root: SafariBookmarkItem,
        format: str,
        only_children=False,
        json=False,
    ):
        if json:
            self.output.write(root.json())
        elif only_children:
            self._render_children(root, format=format)
        else:
            self._render_item(root, format=format)

    def list(
        self,
        path: List[str] = [],
        format: Optional[str] = None,
        simple_format=False,
        json=False,
    ):
        target = self._get_or_walk(path)
        if target is None:
            raise ValueError("Target not found")
        if simple_format:
            format = SIMPLE_FORMAT
        elif format is None:
            format = DEFAULT_LIST_FORMAT
        self._render(target, only_children=target.is_folder, format=format, json=json)

    def add(
        self,
        title: Optional[str],
        uuid: Optional[str] = None,
        url: Optional[str] = None,
        path: List[str] = [],
        list=False,
    ):
        target = self._get_or_walk(path)
        if target is None or not target.is_folder:
            raise ValueError("Invalid destination")
        if list:
            if url:
                raise ValueError("URL is not supported by lists")
            if not title:
                raise ValueError("Title is required")
            target.add_folder(title=title, id=uuid)
        elif url is None:
            raise ValueError("URL is required")
        else:
            target.add_bookmark(url=url, id=uuid, title=title)
        self._save()

    def remove(self, path: List[str]):
        target = self._get_or_walk(path)
        if target is None:
            raise ValueError("Target not found")
        if parent := target.parent:
            parent.remove(target)
        self._save()

    def move(self, path: List[str], to: List[str] = []):
        target = self._get_or_walk(path)
        if target is None:
            raise ValueError("Target not found")
        if not to:
            raise ValueError("Missing destination")
        dest = self._get_or_walk(to)
        if dest is None or not dest.is_folder:
            raise ValueError("Invalid destination")
        dest.append(target)
        self._save()

    def edit(
        self,
        path: List[str],
        title: Optional[str] = None,
        url: Optional[str] = None,
    ):
        target = self._get_or_walk(path)
        if target is None:
            raise ValueError("Target not found")
        if title is not None:
            target.title = title
        if url is not None:
            target.url = url
            if target.is_bookmark:
                target.url = url
            else:
                raise ValueError("Cannot update target url")
        self._save()

    def empty(self, path: List[str]):
        target = self._get_or_walk(path)
        if target is None:
            raise ValueError("Target not found")
        if not target.is_folder:
            raise ValueError("Target is not a list")
        target.empty()
        self._save()


def generate_colors(output: IO) -> dict[str, str]:
    if supports_colors(output):
        return {name: "\033[%dm" % code for name, code in COLORS.items()}
    else:
        return {name: "" for name in COLORS.keys()}


def supports_colors(tty: IO) -> bool:
    if "ANSI_COLORS_DISABLED" in os.environ:
        return False
    if "NO_COLOR" in os.environ:
        return False
    if "FORCE_COLOR" in os.environ:
        return True
    if os.environ.get("TERM") == "dumb":
        return False
    if not hasattr(tty, "fileno"):
        return False
    try:
        return os.isatty(tty.fileno())
    except UnsupportedOperation:
        return tty.isatty()
